#!/usr/bin/perl -- # -*- Perl -*-
# Version 1.0  GS 17.12.01 -> first Version
# Version 1.1  GS 18.12.01 -> CC Definition or Definition placed as first line in annotation
# Version 1.2  GS 18.12.01 -> Concatenation of "Type" if the specific DataType do not have this required extension.
# Version 1.3  GS 08.08.02 -> Improvements
# Version 1.4  ndw 26 Aug 2002 -> Hacking
# Version 2.0  GS 12.11.02 -> Completely Restructured
# Version 2.1  GS 14.01.02 -> Bugfix: Generation of annotation.
# Version 2.2  GS 15.01.02 -> Global defined elements.
# Version 2.3  GS 16.12.02 -> Complex types for ABIEs and BBIEs without qualifiers for object class terms.
#
# Packages
#
use strict;
use Spreadsheet::ParseExcel;
use XML::Parser::PerlSAX;
use XML::writer;
use IO;
#
# Global variables
#
my $usage = "Usage: $0 configFile.xml excelFile.xls outFile.xsd\n";
my $parser = XML::Parser::PerlSAX->new( Handler => configHandler->new() );

my( $configFile, $xlsFile, $xsdOutfile, $rootElement );	# File from command line
my( %hConfTab, %hConfCol, %hConfNs, %ns ) = ();
my( %hCCs, %hAggregates, %hAssociations, %hBasics, %hAnnotations ) = ();
my( @aAnnotationSeq ) = ();
#
# Global definitions
#
push( @aAnnotationSeq, "CC_DEN" );
push( @aAnnotationSeq, "BIE_DEN" );
$hConfCol{CC_DEN}{name} = "Dictionary Entry Name";
$hConfCol{BIE_DEN}{name} = "Dictionary Entry Name";

#
# Reading command line
#
while (@ARGV) {
	my $param = shift @ARGV;
	if ($param =~ /\.xml$/) {
		$configFile = $param;
	} elsif ($param =~ /\.xls$/) {
		$xlsFile = $param;
	} elsif ($param =~ /\.xsd$/) {
		$xsdOutfile = $param;
	} else {
		$rootElement = $param;
	}
}

die $usage if !defined($xlsFile) || ! -f $xlsFile;
die $usage if !defined($configFile) || ! -f $configFile;
die $usage if !defined($xsdOutfile);
print STDERR "Loading $xlsFile...\n";

#
# Initialising of input (excel spreadsheet) and and output (XML Schema)
#
my $oExcel = new Spreadsheet::ParseExcel;
my $oXlsFile = $oExcel->Parse($xlsFile);
my $output = new IO::File(">$xsdOutfile");
# 
# Parsing of configuration XML-file and Excel-Spreadsheet
#
$parser->parse( Source => {SystemId => $configFile} );
excel_parser ($oXlsFile, $hConfTab{ExcelTable}{data} );

#
# Initialising of common namespaces
#
$ns{XSD} = "http://www.w3.org/2001/XMLSchema";
$hConfNs{XHTML}{ns} = "http://www.w3.org/2001/XHTML";
$hConfNs{XHTML}{prefix} = "xhtml";
#
# Writing XML-Schema
#
my $writer = new XML::Writer( NAMESPACES => 1, 
										OUTPUT => $output, 
										DATA_MODE => 'true',
										DATA_INDENT => 2,
										PREFIX_MAP => { $ns{XSD} => 'xsd' });
# Initialising of further namespaces
foreach my $keyNs ( keys %hConfNs ) {
	if( $keyNs ne "XSD" ){
		$ns{$keyNs} = $hConfNs{$keyNs}{ns};
		$writer->addPrefix( $ns{$keyNs}, $hConfNs{$keyNs}{prefix});
		$hConfNs{$keyNs}{attr} = "xmlns:$hConfNs{$keyNs}{prefix}";
	}
}
#
# Writing XML-Schema output
#
$writer->xmlDecl( 'UTF-8' );
$writer->comment( "26 Aug 2002 - source: $xlsFile" );
$writer->comment( "Created by: Gunther Stuhec, SAP AG" );
$writer->startTag( [$ns{XSD}, "schema"], 
						"xmlns:xsd" => $ns{XSD},
						$hConfNs{TARGET}{attr} => $hConfNs{TARGET}{ns},
						$hConfNs{CCTS}{attr} => $hConfNs{CCTS}{ns},
						$hConfNs{CCT}{attr} => $hConfNs{CCT}{ns},
						$hConfNs{CC}{attr} => $hConfNs{CC}{ns},
						$hConfNs{BIE}{attr} => $hConfNs{BIE}{ns} );
foreach my $keyNs ( keys %hConfNs ) {
	if( $keyNs ne "TARGET" and $hConfNs{$keyNs}{loc} ne "" ) {
		$writer->emptyTag( [$ns{XSD}, "import"],
								"namespace" => $hConfNs{$keyNs}{ns},	
								"schemaLocation" => $hConfNs{$keyNs}{loc});
	}
}						
out_aggregate_components();
out_basic_components();

# Creation of the end of xml-schema
if ($rootElement) {
	print STDERR "Outputting top-level element: $rootElement\n";
	$writer->emptyTag( [$ns{XSD}, "element"],
							"name" => $rootElement,	
							"type" => "${rootElement}Details");
} else {
    print STDERR "Warning: No top-level elements defined!\n";
}
$writer->endTag();
$writer->end();

#====================================================================
# FUNCTIONS
#====================================================================

#
# Function: excel_parser
# Parsing the specific excel spreed sheat according the configuration file.
#
sub excel_parser ()
{
	my( $oXlsFile ) = shift;
	my( $iSheet ) = shift;
	my( $iRow, $iCol, $oWorkSheet, $oWorkCol );
	my( $sCellContent, $sAggregateType, $y, $sAggObject );
	my( $i ) = 0;
	
	$oWorkSheet = $oXlsFile->{Worksheet}[$iSheet];
	
	for( my $iRow = $hConfTab{ExcelTable}{lineOfBegin}; defined $oWorkSheet->{MaxRow} && $iRow <= $oWorkSheet->{MaxRow}; $iRow++) {
		%hCCs = ();
		$i++;
		$hCCs{UID} = sprintf('UBL%06d', $i);
		for(my $iCol = $oWorkSheet->{MinCol} ; defined $oWorkSheet->{MaxCol} && $iCol <= $oWorkSheet->{MaxCol} ; $iCol++) {
			$oWorkCol = $oWorkSheet->{Cells}[$iRow][$iCol];
			$sCellContent = $oWorkCol->Value if($oWorkCol);
			foreach my $keyColumn ( keys %hConfCol ) {
				if( $iCol == $hConfCol{$keyColumn}{data} && $sCellContent ne "")	{		
					$hCCs{$keyColumn} = $sCellContent;
					$sCellContent = ();
				}
			}
		}
		#
		# Selection of aggregates or basics
		#
		if( $hCCs{CoreComponentCharacteristic} eq "ACC" or $hCCs{CoreComponentCharacteristic} eq "ABIE" ) {
			$y = 0;
			$sAggObject = $hCCs{ObjectClassTerm};
			$sAggObject =~ s/(\w+)/\u\L$1/g;
		} elsif( $hCCs{CoreComponentCharacteristic} eq "BCC" or $hCCs{CoreComponentCharacteristic} eq "BBIE" or $hCCs{CoreComponentCharacteristic} eq "ASCC" or $hCCs{CoreComponentCharacteristic} eq "ASBIE") {
			$y++;
		}
		$sAggregateType = create_list( $hCCs{CoreComponentCharacteristic}, $y, $sAggregateType, $sAggObject );						
	}
}

sub create_list ()
{
	my $sType = shift;
	my( $sCounter ) = shift;
	my( $sAggregateType ) = shift;
	my( $sAggObject ) = shift;
	my( @aDENParameters );
	my $sObjectClass;

	#
	# Definition of Business Information Entities --> Elements
	#
	@aDENParameters = ( 'BusinessProcess', 'Region', 'OfficialConstraints', 'Product', 'Industry', 'Role', 'SupportingRole', 'SystemConstraints', 'QualifierTerm_ObjectClassTerm', 'ObjectClassTerm', 'PropertyTerm', 'RepresentationTerm' );
	$hCCs{BIE_DEN} = create_dictionary_entry_name( @aDENParameters, \%hCCs );
	$hCCs{BIEName} = create_ubl_name( $hCCs{BIE_DEN}, $sAggObject, $hCCs{CoreComponentCharacteristic}, "TAG" );
	$hAnnotations{$hCCs{UID}}{BIE_DEN} = $hCCs{BIE_DEN};	
	#
	# Definition of Core Components --> Complex Types
	#
	@aDENParameters = ( 'ObjectClassTerm', 'PropertyTerm', 'RepresentationTerm' );
	$hCCs{CC_DEN} = create_dictionary_entry_name( @aDENParameters, \%hCCs );
	$hCCs{CCName} = create_ubl_name( $hCCs{CC_DEN}, $sAggObject, $hCCs{CoreComponentCharacteristic} );
	$hAnnotations{$hCCs{UID}}{CC_DEN} = $hCCs{CC_DEN};
	#
	# Further information for annotation
	#
	$hAnnotations{$hCCs{UID}}{CoreComponentCharacteristic} = $hCCs{CoreComponentCharacteristic};
	#
	# Occurence definition
	#
	($hCCs{MinOccurs}, $hCCs{MaxOccurs}) = min_max_result( $hCCs{Occurrence}, '0', '1' ) if( $hCCs{Occurrence} );
	($hCCs{MinLength}, $hCCs{MaxLength}, $hCCs{FixLength}) = min_max_result( $hCCs{Length} ) if( $hCCs{Length} );

	foreach my $ComponentKey (keys %hCCs) {
		if( $sType eq "ACC" or $sType eq "ABIE" ) {
			$sAggregateType = $hCCs{CCName};		
			$hAggregates{$sAggregateType}{$ComponentKey} = $hCCs{$ComponentKey};			
		} elsif( $sType eq "BCC" or $sType eq "BBIE" ) {
			$hAssociations{$sAggregateType}{$sCounter}{$ComponentKey} = $hCCs{$ComponentKey};
			$hBasics{$hCCs{CCName}}{$ComponentKey} = $hCCs{$ComponentKey};
			if( $ComponentKey eq "RepresentationTerm" ) {
				$hBasics{$hCCs{CCName}}{BaseType} = $hCCs{$ComponentKey};	
				$hBasics{$hCCs{CCName}}{BaseType} =~ s/(\w+)/\u\L$1/g;
				$hBasics{$hCCs{CCName}}{BaseType} =~ s/ //g;
				$hBasics{$hCCs{CCName}}{BaseType} = "$hBasics{$hCCs{CCName}}{BaseType}Type";
			}
		} elsif( $sType eq "ASCC" or $sType eq "ASBIE" ) {
			$hAssociations{$sAggregateType}{$sCounter}{$ComponentKey} = $hCCs{$ComponentKey};
		}
		$hAnnotations{$hCCs{UID}}{$ComponentKey} = $hCCs{$ComponentKey} if( $hConfCol{$ComponentKey}{name} );
	}
	return $sAggregateType;
}

#
# Creation of Aggregates
#
sub out_aggregate_components () 
{
	my( $sType ) = "";

	foreach my $AccKey (sort keys %hAggregates) {
		$writer->startTag( [$ns{XSD}, "complexType"],
								"name" => $hAggregates{$AccKey}{CCName},
								"id" => $hAggregates{$AccKey}{UID} );
#		out_annotation( $AccKey );
		out_ccts_info( $hAggregates{$AccKey}{UID}, "CC" );
		$writer->startTag( [$ns{XSD}, "sequence"] );
		foreach my $AccElementKey (sort { $a <=> $b } keys %{$hAssociations{$AccKey}}) {
			$sType = "";
			$writer->startTag( [$ns{XSD}, "element" ],
									"name" => $hAssociations{$AccKey}{$AccElementKey}{BIEName},
									"type" => $hAssociations{$AccKey}{$AccElementKey}{CCName},
									"id" => $hAssociations{$AccKey}{$AccElementKey}{UID},
									"minOccurs" => $hAssociations{$AccKey}{$AccElementKey}{MinOccurs},
									"maxOccurs" => $hAssociations{$AccKey}{$AccElementKey}{MaxOccurs});																		
#			foreach my $ElementInfoKey (keys %{${$hAggregates{$AccKey}}{$AccElementKey}}) {
#				print "\t\t Info Key: $ElementInfoKey\n";
#			}
# 			out_annotation( $hAssociations{$AccKey}{$AccElementKey}{UID} );
			out_ccts_info( $hAssociations{$AccKey}{$AccElementKey}{UID}, "BIE" );
			$writer->endTag();
		}
		$writer->endTag();
		$writer->endTag();
	}
}

#
# Function: Creation of Basics
#
sub out_basic_components ()
{
	foreach my $BccKey (sort keys %hBasics) {
		my $sId = "";
		my $sType = "";
		$writer->startTag( [$ns{XSD}, "complexType"],
								"name" => $hBasics{$BccKey}{CCName},
								"id" => $hBasics{$BccKey}{UID} );
		$writer->startTag( [$ns{XSD}, "simpleContent"] );
		$writer->startTag( [$ns{XSD}, "restriction"],
								"base" => "$hConfNs{CCT}{prefix}:$hBasics{$BccKey}{BaseType}" );
# 		out_annotation( $BccKey );
		out_ccts_info( $hBasics{$BccKey}{UID}, "CC" );
		$writer->endTag();
		$writer->endTag();
		$writer->endTag();
	}
}

#
# Function: Creation of annotation
#
sub out_annotation ()
{
	my $sUID = shift;
	my $i;
	my ($sDocuValue, $sClassValue) = "";

	$writer->startTag( [$ns{XSD}, "annotation"] );
	$writer->startTag( [$ns{XSD}, "documentation" ] );
	foreach my $keyAnnotation ( @aAnnotationSeq ) {
		if( $hAnnotations{$sUID}{$keyAnnotation} ne "" ) {
			$writer->startTag( [$ns{XHTML}, "div"],
									"class" => $hConfCol{$keyAnnotation}{name} );
			$writer->startTag( [$ns{XHTML}, "p"] );
			$writer->characters( $hAnnotations{$sUID}{$keyAnnotation} );
			$writer->endTag();
			$writer->endTag();
		}
	}
	$writer->endTag();
	$writer->endTag();
}

#
# Function: Creation of annotation
#
sub out_ccts_info ()
{
	my $sUID = shift;
	my $sType = shift;
	my( $i, $sDEN, $sDefinition );
	my( $sDocuValue, $sClassValue ) = "";
	
	if ( $sType eq "BIE" ) {
		$sDEN = $hAnnotations{$sUID}{BIE_DEN};
		$sDefinition = $hAnnotations{$sUID}{Definition};
	} elsif ( $sType eq "CC" ) {
		$sDEN = $hAnnotations{$sUID}{CC_DEN};
		$sDefinition = "";
	}
	
	$writer->startTag( [$ns{XSD}, "annotation"] );
	$writer->startTag( [$ns{XSD}, "documentation" ] );
	$writer->emptyTag( [$ns{CCTS}, $hAnnotations{$sUID}{CoreComponentCharacteristic}],
							"dictionaryEntryName" => $sDEN,
							"definition" => $sDefinition,
							"qualifierTermObjectClassTerm" => $hAnnotations{$sUID}{QualifierTerm_ObjectClassTerm},
							"objectClassTerm" => $hAnnotations{$sUID}{ObjectClassTerm},
							"qualifierTerm_PropertyTerm" => $hAnnotations{$sUID}{QualifierTerm_PropertyTerm},
							"propertyTerm" => $hAnnotations{$sUID}{PropertyTerm},
							"representationTerm" => $hAnnotations{$sUID}{RepresentationTerm} );
							
#	$writer->emptyTag( [$ns{CCTS}, "Context"],
#							"businessProcess" => $hAnnotations{$sUID}{BusinessProcess},
#							"region" => $hAnnotations{$sUID}{Region},
#							"officialConstraints" => $hAnnotations{$sUID}{OfficialConstraints},
#							"product" => $hAnnotations{$sUID}{Product},
#							"industry" => $hAnnotations{$sUID}{Industry},
#							"role" => $hAnnotations{$sUID}{Role},
#							"supportingRole" => $hAnnotations{$sUID}{SupportingRole},
#							"systemConstraints" => $hAnnotations{$sUID}{SystemConstraints} );
	$writer->endTag();
	$writer->endTag();
}

#
# Function of definining min-value and max-value
#
sub min_max_result () 
{
	my $sMinMax = shift;
	my $sDefMin = shift;
	my $sDefMax = shift;
	my $sMin = $sMinMax;
	my $sMax = $sMinMax;
	
	if( $sMin =~ m/^[0-9].*/ ) {
		$sMin =~ s/\.\.[0-9n]*$//g;
	} else {
		$sMin = $sDefMin;
	}
	if( $sMax =~ m/..[0-9].*$/ ) {
		$sMax =~ s/^[0-9]*\.\.//g;
	} elsif( $sMax =~ m/..n$/ ) {
		$sMax = "unbounded";
	} else {
		$sMax = $sDefMax;
	}
	return( $sMin, $sMax );
}

#
# Function: Creation of the dictionary entry name
#
sub create_dictionary_entry_name ()
{
	my @aDENParameters = @_;
	my ($hCCs);
	my $sDEName;
	my $sCCPart;
	my $sPropertyTerm;
	my $sPropertyLastWord;
	my $sRepresentationTerm;
	my $sComponentInfo;

	foreach $sCCPart ( @aDENParameters ) {
		$sComponentInfo = $hCCs{$sCCPart};
		$sComponentInfo =~ s/\*//g;
		$sComponentInfo =~ s/(\w+)/\u\L$1/g;
		$sComponentInfo =~ s/'S/'s/g;
		if( $sCCPart =~ /^Context_.*/ and $sComponentInfo ne '') {
			$sDEName = "$sDEName$sComponentInfo ";
		}
		if( $sCCPart =~ /^QualifierTerm_.*/ and $sComponentInfo ne '') {
			$sComponentInfo =~ s/^[()]//g;
			$sDEName = "${sDEName}${sComponentInfo}_ " if( $sComponentInfo !~ s/[()]//g );
			}
		if( $sCCPart eq 'ObjectClassTerm' and $sComponentInfo ne '') {
			$sDEName = "${sDEName}${sComponentInfo}. ";
			}
		if( $sCCPart eq 'PropertyTerm' and $sComponentInfo ne '' ) {
			$sPropertyTerm = $sComponentInfo if( $sComponentInfo !~ s/[()]//g );
			}
		if( $sCCPart eq 'RepresentationTerm' and $sComponentInfo ne '') {
			$sRepresentationTerm = $sComponentInfo;
			if( $sPropertyTerm ne '' ) {
				$sPropertyLastWord = $sPropertyTerm;
				$sPropertyLastWord =~ s/^.* //g;
				if( $sPropertyLastWord eq $sRepresentationTerm or $sPropertyLastWord eq 'Identification' or $sPropertyLastWord eq 'Id' ) {
					$sPropertyTerm =~ s/$sPropertyLastWord$//g;
				}
				if( $sPropertyTerm eq '' or $sPropertyTerm eq $sRepresentationTerm ) {
					$sDEName = "$sDEName$sRepresentationTerm";
				} else {
					$sPropertyTerm =~ s/ $//g;
					$sDEName = "$sDEName$sPropertyTerm. $sRepresentationTerm";
				}
			} else {
				$sDEName = "$sDEName$sRepresentationTerm";
			}
		}
	}
	return $sDEName; 
}

#
# Function: Creation of the UBL Tage name.
#
sub create_ubl_name ()
{
	my $sUBLName = shift;
	my $sAggObject = shift;
	my $sCCType = shift;
	my $sTagType = shift;
	
	$sAggObject =~ s/^_ //g;
	$sUBLName =~ s/'//g;
	
	if( $sTagType eq 'TAG' ) {
		$sUBLName =~ s/^.*?\. //g if( $sUBLName =~ /$sAggObject.*/g and $sCCType ne 'ABIE' and $sCCType ne 'ACC' );
		$sUBLName =~ s/\. Detail$//g;
		$sUBLName =~ s/\. Details$//g;
	}
	elsif( $sCCType eq 'ASBIE' and $sCCType ne 'ASCC' ) {
		$sUBLName =~ s/^.*?\. //g;
		$sUBLName = "$sUBLName. Details";
	}
	$sUBLName =~ s/\. Text$//g;
	$sUBLName =~ s/\Uniform Resource. Identifier$/.URI/g;
	$sUBLName =~ s/Identifier$/.ID/g;
	$sUBLName =~ s/[_. ]//g;
	$sUBLName =~ s/ //g;
	return $sUBLName;
}

# PACKAGES
#====================================================================
#
# PACKAGE configHandler
# Handling xml based config file
# $parser->parse( Source => {SystemId => $configFile} );
#
package configHandler;
#
my $configElement = ConfigElementHandler->new();
my $sNamespaceKey;
my @aElementStack;
my $i;


#
# initialize the handler package
#
sub new {
    my $type = shift;
    return bless {}, $type;
}

#
# handle a start-of-element event: output start tag and attributes
#
sub start_element {
	my ( $self, $properties ) = @_;
	push( @aElementStack, $properties->{'Name'} );
	$i = (scalar @aElementStack)-1;
	
	my %hAttr = %{$properties->{'Attributes'}};
	if ( $aElementStack[1] eq "Namespace" ) {
		$sNamespaceKey = $hAttr{type} if ( $hAttr{type} );
		$hConfNs{$sNamespaceKey}{prefix} = $hAttr{prefix} if ( $hAttr{prefix} );
		$hConfNs{$sNamespaceKey}{loc} = $hAttr{location} if ( $hAttr{location} );
	} elsif ( $aElementStack[1] eq  "Columns" ) {
		if( $hAttr{name} ) {
			$hConfCol{$aElementStack[$i]}{name} = $hAttr{name};
			push( @aAnnotationSeq, $aElementStack[$i] );
		}
	} elsif ( $i == 1 ) {
		$hConfTab{$aElementStack[$i]}{base} = $hAttr{base} if ( $hAttr{base} );
		$hConfTab{$aElementStack[$i]}{lineOfBegin} = $hAttr{lineOfBegin} if ( $hAttr{lineOfBegin} );
	}
}

#
# handle an end-of-element event: output end tag UNLESS it's from a
# <literal> inside a <programlisting>
#
sub end_element {
	my( $self, $properties ) = @_;
	pop( @aElementStack );
}

#
# handle a character data event
#
sub characters {
	my( $self, $properties ) = @_;
	my $data = $properties->{'Data'};
	if ( $aElementStack[1] eq "Namespace" ) {
		$hConfNs{$sNamespaceKey}{ns} = $data;
	} elsif ( $aElementStack[1] eq  "Columns" ) {
		$hConfCol{$aElementStack[$i]}{data} = $data;		
	} elsif( $aElementStack[$i] and $data ne "\n"){
		$hConfTab{$aElementStack[$i]}{data} = $data;
	}
	$sNamespaceKey = ();
}

#
# PACKAGE ConfigElementHandler
# Handling xml based config file
#
package ConfigElementHandler;
#
# initialize the handler package
#
sub new {
    my $type = shift;
    return bless {}, $type;
}
